with Ada.Unchecked_Conversion;
with Interfaces.C;
with System.Address_To_Access_Conversions;
with System.Storage_Elements;
package Windows is

	pragma Ada_05;
	pragma Suppress(All_Checks);

---- Translated from WINDEF.h, WINNT.h ----

	subtype Pointer is System.Address; -- generic pointer

	-- Ansi String
	subtype ACHAR is Interfaces.C.char;
	type PACHAR is access all ACHAR;
	type PCACHAR is access constant ACHAR;
	subtype PASTR is PACHAR;
	subtype PCASTR is PCACHAR;
	subtype ASTR is Interfaces.C.char_array;
	ANULL: ACHAR renames Interfaces.C.nul;
	package Address_To_PCASTR_Conversions is new System.Address_To_Access_Conversions(ACHAR);

	-- Wide String
	subtype WCHAR is Interfaces.C.wchar_t;
	type PWCHAR is access all WCHAR;
	type PCWCHAR is access constant WCHAR;
	subtype PWSTR is PWCHAR;
	subtype PCWSTR is PCWCHAR;
	subtype WSTR is Interfaces.C.wchar_array;
	WNULL: WCHAR renames Interfaces.C.wide_nul;
	package Address_To_PCWSTR_Conversions is new System.Address_To_Access_Conversions(WCHAR);

	-- String
	subtype TCHAR is ACHAR;
	subtype PTCHAR is PACHAR;
	subtype PCTCHAR is PCACHAR;
	subtype PTSTR is PASTR;
	subtype PCTSTR is PCASTR;
	subtype TSTR is ASTR;
	TNULL : TCHAR renames ANULL;
	package Address_To_PCTSTR_Conversions renames Address_To_PCASTR_Conversions;

	-- 8 bit Unsigned
	subtype BYTE is Interfaces.Integer_8; --new Interfaces.C.unsigned_char;
	type PBYTE is access all BYTE;
	type PCBYTE is access constant BYTE;

	-- 16 bit Signed
	subtype SHORT is Interfaces.Integer_16; --new Interfaces.C.short;
	type PSHORT is access all SHORT;
	type PCSHORT is access constant SHORT;

	-- 16 bit Unsigned
	subtype WORD is Interfaces.Unsigned_16; --new Interfaces.C.unsigned_short;
	type PWORD is access all WORD;
	type PCWORD is access constant WORD;
	subtype USHORT is WORD;
	subtype PUSHORT is PWORD;
	subtype PCUSHORT is PCWORD;

	-- 32 bit Signed
	subtype LONG is Interfaces.Integer_32; --Interfaces.C.long;
	type PLONG is access all LONG;
	type PCLONG is access constant LONG;
	use type LONG;

	-- 32 bit Unsigned
	subtype DWORD is Interfaces.Unsigned_32; --Interfaces.C.unsigned_long;
	type PDWORD is access all DWORD;
	type PCDWORD is access constant DWORD;
	subtype ULONG is DWORD;
	subtype PULONG is PDWORD;
	subtype PCULONG is PCDWORD;
	use type DWORD;

	-- 64 bit signed integer
	subtype LONGLONG is Interfaces.Integer_64;
	type PLONGLONG is access all LONGLONG;
	type PCLONGLONG is access constant LONGLONG;

	-- 64 bit unsigned integer
	subtype DWORDLONG is Interfaces.Unsigned_64;
	type PDWORDLONG is access all DWORDLONG;
	type PCDWORDLONG is access constant DWORDLONG;

	-- Integer
	subtype INT is LONG;
	subtype PINT is PLONG;
	subtype PCINT is PCLONG;

	-- Unsigned
	subtype UINT is ULONG;
	subtype PUINT is PULONG;
	subtype PCUINT is PCULONG;
	
	-- 32 bit Float
	subtype FLOAT  is Interfaces.C.c_float;
	type PFLOAT is access all FLOAT;
	type PCFLOAT is access constant FLOAT;

	-- 64 bit Float
	subtype DOUBLE  is Interfaces.C.double;
	type PDOUBLE is access all DOUBLE;
	type PCDOUBLE is access constant DOUBLE;

	-- Boolean

	type BOOL is new INT;
	type PBOOL is access all BOOL;
	type PCBOOL is access constant BOOL;
	BOOL_FALSE : constant := BOOL'(0);
	BOOL_TRUE  : constant := BOOL'(1);
	function "not" (X: BOOL) return Boolean;
	pragma Inline("not");
	function "+" (X: BOOL) return Boolean;
	pragma Inline("+");

	-- Handle

	type HANDLE is new UINT;
	type PHANDLE is access all HANDLE;
	type PCHANDLE is access constant HANDLE;

	subtype HINSTANCE is HANDLE;
	subtype HWND is HANDLE;
	subtype HICON is HANDLE;
	subtype HBRUSH is HANDLE;
	subtype HCURSOR is HANDLE;
	subtype HMENU is HANDLE;
	subtype HGDIOBJ is HANDLE;

	-- ATOM

	subtype ATOM is WORD;

	-- Point, Size, Rect

	type POINT is record 
		x: LONG;
		y: LONG;
	end record;
	pragma Convention(C, POINT);

	subtype POINTL is POINT;

	type COORD is record
		X: SHORT;
		Y: SHORT;
	end record;
	pragma Convention(C, COORD);

	type COORD_PASS_BY_VALUE is new DWORD;
	function Pass_By_Value is new Ada.Unchecked_Conversion(COORD, COORD_PASS_BY_VALUE);
	function "+" (X: COORD) return COORD_PASS_BY_VALUE renames Pass_By_Value;

	subtype POINTS is COORD;

	type SIZE is record 
		cx: LONG; 
		cy: LONG; 
	end record;
	pragma Convention(C, SIZE);

	subtype SIZEL is SIZE;

	type RECT is record
		Left: LONG; 
		Top: LONG; 
		Right: LONG; 
		Bottom: LONG;
	end record;
	pragma Convention(C, RECT);

	subtype RECTL is RECT;

	type SMALL_RECT is record
		Left  : SHORT; 
		Top   : SHORT; 
		Right : SHORT; 
		Bottom: SHORT;
	end record;
	pragma Convention(C, SMALL_RECT);

	subtype RECTS is SMALL_RECT;

	type PSMALL_RECT is access all SMALL_RECT;
	type PCSMALL_RECT is access constant SMALL_RECT;

	-- parameters
	subtype WPARAM is UINT;
	subtype LPARAM is LONG;
	subtype LRESULT is LONG;
	subtype LCID is DWORD;
	subtype LANGID is WORD;
	subtype PSID is Pointer;
	MAX_PATH : constant := 260;

	GENERIC_READ    : constant := 16#80000000#;
	GENERIC_WRITE   : constant := 16#40000000#;
	GENERIC_EXECUTE : constant := 16#20000000#;
	GENERIC_ALL     : constant := 16#10000000#;

---- WINBASE.h ----

	function GetLastError return DWORD;
	pragma Import(stdcall, GetLastError, "GetLastError");
	
	FORMAT_MESSAGE_ALLOCATE_BUFFER : constant := 16#0100#;
	FORMAT_MESSAGE_IGNORE_INSERTS  : constant := 16#0200#;
	FORMAT_MESSAGE_FROM_STRING     : constant := 16#0400#;
	FORMAT_MESSAGE_FROM_HMODULE    : constant := 16#0800#;
	FORMAT_MESSAGE_FROM_SYSTEM     : constant := 16#1000#;
	FORMAT_MESSAGE_ARGUMENT_ARRAY  : constant := 16#2000#;
	FORMAT_MESSAGE_MAX_WIDTH_MASK  : constant := 16#00FF#;
	
	function FormatMessageA(Flags: DWORD; Source: Pointer; MessageId: DWORD; LanguageId: DWORD; Buffer: access ACHAR; Size: DWORD; Arguments: Pointer) return DWORD; 
	function FormatMessageW(Flags: DWORD; Source: Pointer; MessageId: DWORD; LanguageId: DWORD; Buffer: access WCHAR; Size: DWORD; Arguments: Pointer) return DWORD; 
	function FormatMessage (Flags: DWORD; Source: Pointer; MessageId: DWORD; LanguageId: DWORD; Buffer: access TCHAR; Size: DWORD; Arguments: Pointer) return DWORD renames FormatMessageA;
	pragma Import(stdcall, FormatMessageA, "FormatMessageA");
	pragma Import(stdcall, FormatMessageW, "FormatMessageW");
	
	function GetCurrentThreadId return DWORD;
	pragma Import(stdcall, GetCurrentThreadId, "GetCurrentThreadId");
	
	procedure Sleep(Milliseconds: DWORD);
	pragma Import(stdcall, Sleep, "Sleep");
	
	function CloseHandle(Object: HANDLE) return BOOL;
	pragma Import(stdcall, CloseHandle, "CloseHandle");
	procedure CloseHandle(Object: HANDLE);
	pragma Inline(CloseHandle);
	
	type SECURITY_ATTRIBUTES is record
		Length: DWORD;
		SecurityDescriptor: Pointer;
		InheritHandle: BOOL;
	end record;
	pragma Convention(C, SECURITY_ATTRIBUTES);
	
	type PSECURITY_ATTRIBUTES is access all SECURITY_ATTRIBUTES;
	type PCSECURITY_ATTRIBUTES is access constant SECURITY_ATTRIBUTES;

	INVALID_HANDLE_VALUE : constant := HANDLE'(16#FFFFFFFF#);

	STD_INPUT_HANDLE  : constant := DWORD'(16#FFFFFFF6#);
	STD_OUTPUT_HANDLE : constant := DWORD'(16#FFFFFFF5#);
	STD_ERROR_HANDLE  : constant := DWORD'(16#FFFFFFF4#);

	function GetStdHandle(StdHandle: DWORD) return HANDLE;
	pragma Import(stdcall, GetStdHandle, "GetStdHandle");

	type FILETIME is record
		LowDateTime  : DWORD;
		HighDateTime : DWORD;
	end record;
	pragma Convention(C, FILETIME);

	type WIN32_FIND_DATAA is record
		FileAttributes    : DWORD;
		CreationTime      : FILETIME;
		LastAccessTime    : FILETIME;
		LastWriteTime     : FILETIME;
		FileSizeHigh      : DWORD;
		FileSizeLow       : DWORD;
		Reserved0         : DWORD;
		Reserved1         : DWORD;
		FileName          : Interfaces.C.char_array(0..259);
		AlternateFileName : Interfaces.C.char_array(0..13);
	end record;
	pragma Convention(C, WIN32_FIND_DATAA);

	type WIN32_FIND_DATAW is
		record
			FileAttributes   : DWORD;
			CreationTime     : FILETIME; 
			LastAccessTime   : FILETIME; 
			LastWriteTime    : FILETIME; 
			FileSizeHigh     : DWORD;
			FileSizeLow      : DWORD;
			Reserved0        : DWORD;
			Reserved1        : DWORD;
			FileName         : Interfaces.C.wchar_array(0..259);
			AlternateFileName: Interfaces.C.wchar_array(0..13);
		end record;
	pragma Convention(C, WIN32_FIND_DATAW);

	subtype WIN32_FIND_DATA is WIN32_FIND_DATAA;

	function FindFirstFileA(FileName: PCASTR; FindFileData: access WIN32_FIND_DATAA) return HANDLE;
	function FindFirstFileW(FileName: PCWSTR; FindFileData: access WIN32_FIND_DATAW) return HANDLE;
	function FindFirstFile (FileName: PCTSTR; FindFileData: access WIN32_FIND_DATA ) return HANDLE renames FindFirstFileA;
	pragma Import(stdcall, FindFirstFileA, "FindFirstFileA");
	pragma Import(stdcall, FindFirstFileW, "FindFirstFileW");

	function FindNextFileA(FindFile: HANDLE; FindFileData: access WIN32_FIND_DATAA) return BOOL;
	function FindNextFileW(FindFile: HANDLE; FindFileData: access WIN32_FIND_DATAW) return BOOL;
	function FindNextFile (FindFile: HANDLE; FindFileData: access WIN32_FIND_DATA ) return BOOL renames FindNextFileA;
	pragma Import(stdcall, FindNextFileA, "FindNextFileA");
	pragma Import(stdcall, FindNextFileW, "FindNextFileW");

	function FindClose(FindFile: HANDLE) return BOOL;
	pragma Import(stdcall, FindClose, "FindClose");
	procedure FindClose(FindFile: HANDLE);
	pragma Inline(FindClose);

	function lstrlenA(String: PCASTR) return INT;
	function lstrlenW(String: PCWSTR) return INT;
	function lstrlen (String: PCTSTR) return INT renames lstrlenA;
	pragma Import(stdcall, lstrlenA, "lstrlenA");
	pragma Import(stdcall, lstrlenW, "lstrlenW");

	function lstrcpyA(String1: PASTR; String2: PCASTR) return PASTR;
	function lstrcpyW(String1: PWSTR; String2: PCWSTR) return PWSTR;
	function lstrcpy (String1: PTSTR; String2: PCTSTR) return PTSTR renames lstrcpyA;
	pragma Import(stdcall, lstrcpyA, "lstrcpyA");
	pragma Import(stdcall, lstrcpyW, "lstrcpyW");

---- WINNLS.h ----

	function IsDBCSLeadByte(TestChar: ACHAR) return BOOL;
	pragma Import(stdcall, IsDBCSLeadByte, "IsDBCSLeadByte");

	function MultiByteToWideChar(CodePage : UINT; Flags : DWORD; MultiByteStr : PCASTR; MultiByte : int; WideCharStr : PWSTR; Max : int) return int;
	pragma Import(stdcall, MultiByteToWideChar, "MultiByteToWideChar");

---- WINCONS.h ----

	function SetConsoleTitleA(ConsoleTitle: PCASTR) return BOOL;
	function SetConsoleTitleW(ConsoleTitle: PCWSTR) return BOOL;
	function SetConsoleTitle (ConsoleTitle: PCTSTR) return BOOL renames SetConsoleTitleA;
	pragma Import(stdcall, SetConsoleTitleA, "SetConsoleTitleA");
	pragma Import(stdcall, SetConsoleTitleW, "SetConsoleTitleW");
	procedure SetConsoleTitleA(ConsoleTitle: PCASTR);
	procedure SetConsoleTitleW(ConsoleTitle: PCWSTR);
	procedure SetConsoleTitle (ConsoleTitle: PCTSTR) renames SetConsoleTitleA;
	pragma Inline(SetConsoleTitleA);
	pragma Inline(SetConsoleTitleW);

	FOREGROUND_BLUE                : constant := 16#01#;
	FOREGROUND_GREEN               : constant := 16#02#;
	FOREGROUND_RED                 : constant := 16#04#;
	FOREGROUND_INTENSITY           : constant := 16#08#;
	BACKGROUND_BLUE                : constant := 16#10#;
	BACKGROUND_GREEN               : constant := 16#20#;
	BACKGROUND_RED                 : constant := 16#40#;
	BACKGROUND_INTENSITY           : constant := 16#80#;

	function SetConsoleTextAttribute(ConsoleOutput: HANDLE; Attributes: WORD) return BOOL;
	pragma Import(stdcall, SetConsoleTextAttribute, "SetConsoleTextAttribute");

	type Console_Char_Kind is (Unicode, Ascii);
	type Console_Char(Kind: Console_Char_Kind := Ascii) is
		record
			case Kind is
			when Unicode =>
				UnicodeChar: WCHAR;
			when Ascii =>
				AsciiChar: ACHAR;
			end case;
		end record;
	pragma Convention(C, Console_Char);
	pragma Unchecked_Union(Console_Char);

	type KEY_EVENT_RECORD is
		record
			KeyDown        : BOOL;
			RepeatCount    : WORD;
			VirtualKeyCode : WORD;
			VirtualScanCode: WORD;
			Char           : Console_Char;
			ControlKeyState: DWORD;
		end record;

	type MOUSE_EVENT_RECORD is
		record
			MousePosition  : COORD;
			ButtonState    : DWORD;
			ControlKeyState: DWORD;
			EventFlags     : DWORD;
		end record;

	type WINDOW_BUFFER_SIZE_RECORD is
		record
			Size: COORD;
		end record;

	type MENU_EVENT_RECORD is
		record
			CommandId: UINT;
		end record;

	type FOCUS_EVENT_RECORD is
		record
			SetFocus: BOOL;
		end record;

	KEY_EVENT                : constant := 16#0001#;
	MOUSE_EVENT              : constant := 16#0002#;
	WINDOW_BUFFER_SIZE_EVENT : constant := 16#0004#;
	MENU_EVENT               : constant := 16#0008#;
	FOCUS_EVENT              : constant := 16#0010#;

	type Console_Event(Kind: WORD := KEY_EVENT) is
		record
			case Kind is
			when KEY_EVENT =>
				KeyEvent: KEY_EVENT_RECORD;
			when MOUSE_EVENT =>
				MouseEvent: MOUSE_EVENT_RECORD;
			when WINDOW_BUFFER_SIZE_EVENT =>
				WindowBufferSizeEvent: WINDOW_BUFFER_SIZE_RECORD;
			when MENU_EVENT =>
				MenuEvent: MENU_EVENT_RECORD;
			when others =>
				FocusEvent: FOCUS_EVENT_RECORD;
			end case;
		end record;
	pragma Unchecked_Union(Console_Event);
	pragma Convention(C, Console_Event);

	type INPUT_RECORD is
		record
			EventType: WORD;
			Event: Console_Event;
		end record;
	pragma Convention(C, INPUT_RECORD);
	
	type PINPUT_RECORD is access all INPUT_RECORD;

	function GetNumberOfConsoleInputEvents(ConsoleInput: HANDLE; NumberOfEvents: PDWORD) return BOOL;
	pragma Import(stdcall, GetNumberOfConsoleInputEvents, "GetNumberOfConsoleInputEvents");

	function ReadConsoleInputA(ConsoleInput: HANDLE; Buffer: PINPUT_RECORD; Length: DWORD; NumberOfEventsRead: PDWORD) return BOOL;
	function ReadConsoleInputW(ConsoleInput: HANDLE; Buffer: PINPUT_RECORD; Length: DWORD; NumberOfEventsRead: PDWORD) return BOOL;
	function ReadConsoleInput (ConsoleInput: HANDLE; Buffer: PINPUT_RECORD; Length: DWORD; NumberOfEventsRead: PDWORD) return BOOL renames ReadConsoleInputA;
	pragma Import(stdcall, ReadConsoleInputA, "ReadConsoleInputA");
	pragma Import(stdcall, ReadConsoleInputW, "ReadConsoleInputW");

	type CHAR_INFO is
		record
			Char: Console_Char;
			Attributes: WORD;
		end record;
	pragma Convention(C, CHAR_INFO);
	
	type PCHAR_INFO is access all CHAR_INFO;
	type PCCHAR_INFO is access constant CHAR_INFO;

	function WriteConsoleOutputA(ConsoleOutput: HANDLE; Buffer: PCCHAR_INFO; BufferSize: COORD_PASS_BY_VALUE; BufferCoord: COORD_PASS_BY_VALUE; WriteRegion: PSMALL_RECT) return BOOL;
	function WriteConsoleOutputW(ConsoleOutput: HANDLE; Buffer: PCCHAR_INFO; BufferSize: COORD_PASS_BY_VALUE; BufferCoord: COORD_PASS_BY_VALUE; WriteRegion: PSMALL_RECT) return BOOL;
	function WriteConsoleOutput (ConsoleOutput: HANDLE; Buffer: PCCHAR_INFO; BufferSize: COORD_PASS_BY_VALUE; BufferCoord: COORD_PASS_BY_VALUE; WriteRegion: PSMALL_RECT) return BOOL renames WriteConsoleOutputA;
	pragma Import(stdcall, WriteConsoleOutputA, "WriteConsoleOutputA");
	pragma Import(stdcall, WriteConsoleOutputW, "WriteConsoleOutputW");
	procedure WriteConsoleOutputA(ConsoleOutput: HANDLE; Buffer: PCCHAR_INFO; BufferSize: COORD_PASS_BY_VALUE; BufferCoord: COORD_PASS_BY_VALUE; WriteRegion: PSMALL_RECT);
	procedure WriteConsoleOutputW(ConsoleOutput: HANDLE; Buffer: PCCHAR_INFO; BufferSize: COORD_PASS_BY_VALUE; BufferCoord: COORD_PASS_BY_VALUE; WriteRegion: PSMALL_RECT);
	procedure WriteConsoleOutput (ConsoleOutput: HANDLE; Buffer: PCCHAR_INFO; BufferSize: COORD_PASS_BY_VALUE; BufferCoord: COORD_PASS_BY_VALUE; WriteRegion: PSMALL_RECT) renames WriteConsoleOutputA;
	pragma Inline(WriteConsoleOutputA);
	pragma Inline(WriteConsoleOutputW);

	type CONSOLE_CURSOR_INFO is
		record
			Size: DWORD; 
			Visible: BOOL; 
		end record;
	pragma Convention(C, CONSOLE_CURSOR_INFO);
	type PCCONSOLE_CURSOR_INFO is access constant CONSOLE_CURSOR_INFO;

	function SetConsoleCursorInfo(ConsoleOutput: HANDLE; ConsoleCursorInfo: PCCONSOLE_CURSOR_INFO) return BOOL;
	pragma Import(stdcall, SetConsoleCursorInfo, "SetConsoleCursorInfo");
	procedure SetConsoleCursorInfo(ConsoleOutput: HANDLE; ConsoleCursorInfo: PCCONSOLE_CURSOR_INFO);
	pragma Inline(SetConsoleCursorInfo);

	CONSOLE_TEXTMODE_BUFFER : constant := 1;

	function CreateConsoleScreenBuffer(DesiredAccess: DWORD; ShareMode: DWORD; SecurityAttributes: PCSECURITY_ATTRIBUTES; Flags: DWORD; ScreenBufferData: Pointer) return HANDLE;
	pragma Import(stdcall, CreateConsoleScreenBuffer, "CreateConsoleScreenBuffer");

	function SetConsoleActiveScreenBuffer(ConsoleOutput: HANDLE) return BOOL;
	pragma Import(stdcall, SetConsoleActiveScreenBuffer, "SetConsoleActiveScreenBuffer");
	procedure SetConsoleActiveScreenBuffer(ConsoleOutput: HANDLE);
	pragma Inline(SetConsoleActiveScreenBuffer);

	type CONSOLE_SCREEN_BUFFER_INFO is
		record
			Size             : COORD; 
			CursorPosition   : COORD; 
			Attributes       : WORD; 
			Window           : SMALL_RECT; 
			MaximumWindowSize: COORD; 
		end record;
	pragma Convention(C, CONSOLE_SCREEN_BUFFER_INFO);

	function SetConsoleWindowInfo(ConsoleOutput: HANDLE; Absolute: BOOL; ConsoleWindow: PCSMALL_RECT) return BOOL; 
	pragma Import(stdcall, SetConsoleWindowInfo, "SetConsoleWindowInfo");
	procedure SetConsoleWindowInfo(ConsoleOutput: HANDLE; Absolute: BOOL; ConsoleWindow: PCSMALL_RECT);
	pragma Inline(SetConsoleWindowInfo);

	function SetConsoleScreenBufferSize(ConsoleOutput: HANDLE; Size: COORD_PASS_BY_VALUE) return BOOL;
	pragma Import(stdcall, SetConsoleScreenBufferSize, "SetConsoleScreenBufferSize");
	procedure SetConsoleScreenBufferSize(ConsoleOutput: HANDLE; Size: COORD_PASS_BY_VALUE);
	pragma Inline(SetConsoleScreenBufferSize);

	ENABLE_PROCESSED_INPUT    : constant := DWORD'(1); --  Input Mode flags
	ENABLE_LINE_INPUT         : constant := DWORD'(2);
	ENABLE_ECHO_INPUT         : constant := DWORD'(4);
	ENABLE_WINDOW_INPUT       : constant := DWORD'(8);
	ENABLE_MOUSE_INPUT        : constant := DWORD'(16);
	ENABLE_PROCESSED_OUTPUT   : constant := DWORD'(1); -- Output Mode flags
	ENABLE_WRAP_AT_EOL_OUTPUT : constant := DWORD'(2);

	function GetConsoleMode(ConsoleHandle: HANDLE; Mode: PDWORD) return BOOL;
	pragma Import(stdcall, GetConsoleMode, "GetConsoleMode");

	function SetConsoleMode(ConsoleHandle: HANDLE; Mode: DWORD) return BOOL;
	pragma Import(stdcall, SetConsoleMode, "SetConsoleMode");

---- WINUSER.h ----

	WM_NULL                  : constant := 16#0000#;
	WM_CREATE                : constant := 16#0001#;
	WM_DESTROY               : constant := 16#0002#;
	WM_MOVE                  : constant := 16#0003#;
	WM_SIZE                  : constant := 16#0005#;
	WM_ACTIVATE              : constant := 16#0006#;
	WM_SETFOCUS              : constant := 16#0007#;
	WM_KILLFOCUS             : constant := 16#0008#;
	WM_ENABLE                : constant := 16#000A#;
	WM_SETREDRAW             : constant := 16#000B#;
	WM_SETTEXT               : constant := 16#000C#;
	WM_GETTEXT               : constant := 16#000D#;
	WM_GETTEXTLENGTH         : constant := 16#000E#;
	WM_PAINT                 : constant := 16#000F#;
	WM_CLOSE                 : constant := 16#0010#;
	WM_QUERYENDSESSION       : constant := 16#0011#;
	WM_QUIT                  : constant := 16#0012#;
	WM_QUERYOPEN             : constant := 16#0013#;
	WM_ERASEBKGND            : constant := 16#0014#;
	WM_SYSCOLORCHANGE        : constant := 16#0015#;
	WM_ENDSESSION            : constant := 16#0016#;
	WM_SYSTEMERROR           : constant := 16#0017#;
	WM_SHOWWINDOW            : constant := 16#0018#;
	WM_CTLCOLOR              : constant := 16#0019#;
	WM_WININICHANGE          : constant := 16#001A#;
	WM_SETTINGCHANGE         : constant := WM_WININICHANGE;
	WM_DEVMODECHANGE         : constant := 16#001B#;
	WM_ACTIVATEAPP           : constant := 16#001C#;
	WM_FONTCHANGE            : constant := 16#001D#;
	WM_TIMECHANGE            : constant := 16#001E#;
	WM_CANCELMODE            : constant := 16#001F#;
	WM_SETCURSOR             : constant := 16#0020#;
	WM_MOUSEACTIVATE         : constant := 16#0021#;
	WM_CHILDACTIVATE         : constant := 16#0022#;
	WM_QUEUESYNC             : constant := 16#0023#;
	WM_GETMINMAXINFO         : constant := 16#0024#;
	WM_PAINTICON             : constant := 16#0026#;
	WM_ICONERASEBKGND        : constant := 16#0027#;
	WM_NEXTDLGCTL            : constant := 16#0028#;
	WM_SPOOLERSTATUS         : constant := 16#002A#;
	WM_DRAWITEM              : constant := 16#002B#;
	WM_MEASUREITEM           : constant := 16#002C#;
	WM_DELETEITEM            : constant := 16#002D#;
	WM_VKEYTOITEM            : constant := 16#002E#;
	WM_CHARTOITEM            : constant := 16#002F#;
	WM_SETFONT               : constant := 16#0030#;
	WM_GETFONT               : constant := 16#0031#;
	WM_SETHOTKEY             : constant := 16#0032#;
	WM_GETHOTKEY             : constant := 16#0033#;
	WM_QUERYDRAGICON         : constant := 16#0037#;
	WM_COMPAREITEM           : constant := 16#0039#;
	WM_GETOBJECT             : constant := 16#003D#;
	WM_COMPACTING            : constant := 16#0041#;
	WM_COMMNOTIFY            : constant := 16#0044#;
	WM_WINDOWPOSCHANGING     : constant := 16#0046#;
	WM_WINDOWPOSCHANGED      : constant := 16#0047#;
	WM_POWER                 : constant := 16#0048#;
	WM_COPYDATA              : constant := 16#004A#;
	WM_CANCELJOURNAL         : constant := 16#004B#;
	WM_NOTIFY                : constant := 16#004E#;
	WM_INPUTLANGCHANGEREQUEST: constant := 16#0050#;
	WM_INPUTLANGCHANGE       : constant := 16#0051#;
	WM_TCARD                 : constant := 16#0052#;
	WM_HELP                  : constant := 16#0053#;
	WM_USERCHANGED           : constant := 16#0054#;
	WM_NOTIFYFORMAT          : constant := 16#0055#;
	WM_CONTEXTMENU           : constant := 16#007B#;
	WM_STYLECHANGING         : constant := 16#007C#;
	WM_STYLECHANGED          : constant := 16#007D#;
	WM_DISPLAYCHANGE         : constant := 16#007E#;
	WM_GETICON               : constant := 16#007F#;
	WM_SETICON               : constant := 16#0080#;
	WM_NCCREATE              : constant := 16#0081#;
	WM_NCDESTROY             : constant := 16#0082#;
	WM_NCCALCSIZE            : constant := 16#0083#;
	WM_NCHITTEST             : constant := 16#0084#;
	WM_NCPAINT               : constant := 16#0085#;
	WM_NCACTIVATE            : constant := 16#0086#;
	WM_GETDLGCODE            : constant := 16#0087#;
	WM_NCMOUSEMOVE           : constant := 16#00A0#;
	WM_NCLBUTTONDOWN         : constant := 16#00A1#;
	WM_NCLBUTTONUP           : constant := 16#00A2#;
	WM_NCLBUTTONDBLCLK       : constant := 16#00A3#;
	WM_NCRBUTTONDOWN         : constant := 16#00A4#;
	WM_NCRBUTTONUP           : constant := 16#00A5#;
	WM_NCRBUTTONDBLCLK       : constant := 16#00A6#;
	WM_NCMBUTTONDOWN         : constant := 16#00A7#;
	WM_NCMBUTTONUP           : constant := 16#00A8#;
	WM_NCMBUTTONDBLCLK       : constant := 16#00A9#;
	WM_KEYFIRST              : constant := 16#0100#;
	WM_KEYDOWN               : constant := 16#0100#;
	WM_KEYUP                 : constant := 16#0101#;
	WM_CHAR                  : constant := 16#0102#;
	WM_DEADCHAR              : constant := 16#0103#;
	WM_SYSKEYDOWN            : constant := 16#0104#;
	WM_SYSKEYUP              : constant := 16#0105#;
	WM_SYSCHAR               : constant := 16#0106#;
	WM_SYSDEADCHAR           : constant := 16#0107#;
	WM_KEYLAST               : constant := 16#0108#;
	WM_INITDIALOG            : constant := 16#0110#;
	WM_COMMAND               : constant := 16#0111#;
	WM_SYSCOMMAND            : constant := 16#0112#;
	WM_TIMER                 : constant := 16#0113#;
	WM_HSCROLL               : constant := 16#0114#;
	WM_VSCROLL               : constant := 16#0115#;
	WM_INITMENU              : constant := 16#0116#;
	WM_INITMENUPOPUP         : constant := 16#0117#;
	WM_MENUSELECT            : constant := 16#011F#;
	WM_MENUCHAR              : constant := 16#0120#;
	WM_ENTERIDLE             : constant := 16#0121#;
	WM_MENURBUTTONUP         : constant := 16#0122#;
	WM_MENUDRAG              : constant := 16#0123#;
	WM_MENUGETOBJECT         : constant := 16#0124#;
	WM_UNINITMENUPOPUP       : constant := 16#0125#;
	WM_MENUCOMMAND           : constant := 16#0126#;
	WM_CHANGEUISTATE         : constant := 16#0127#;
	WM_UPDATEUISTATE         : constant := 16#0128#;
	WM_QUERYUISTATE          : constant := 16#0129#;
	WM_CTLCOLORMSGBOX        : constant := 16#0132#;
	WM_CTLCOLOREDIT          : constant := 16#0133#;
	WM_CTLCOLORLISTBOX       : constant := 16#0134#;
	WM_CTLCOLORBTN           : constant := 16#0135#;
	WM_CTLCOLORDLG           : constant := 16#0136#;
	WM_CTLCOLORSCROLLBAR     : constant := 16#0137#;
	WM_CTLCOLORSTATIC        : constant := 16#0138#;
	WM_MOUSEFIRST            : constant := 16#0200#;
	WM_MOUSEMOVE             : constant := 16#0200#;
	WM_LBUTTONDOWN           : constant := 16#0201#;
	WM_LBUTTONUP             : constant := 16#0202#;
	WM_LBUTTONDBLCLK         : constant := 16#0203#;
	WM_RBUTTONDOWN           : constant := 16#0204#;
	WM_RBUTTONUP             : constant := 16#0205#;
	WM_RBUTTONDBLCLK         : constant := 16#0206#;
	WM_MBUTTONDOWN           : constant := 16#0207#;
	WM_MBUTTONUP             : constant := 16#0208#;
	WM_MBUTTONDBLCLK         : constant := 16#0209#;
	WM_MOUSEWHEEL            : constant := 16#020A#;
	WM_MOUSELAST             : constant := 16#020A#;
	WM_PARENTNOTIFY          : constant := 16#0210#;
	WM_ENTERMENULOOP         : constant := 16#0211#;
	WM_EXITMENULOOP          : constant := 16#0212#;
	WM_NEXTMENU              : constant := 16#0213#;
	WM_SIZING                : constant := 16#32#;
	WM_CAPTURECHANGED        : constant := 16#33#;
	WM_MOVING                : constant := 16#34#;
	WM_POWERBROADCAST        : constant := 16#36#;
	WM_DEVICECHANGE          : constant := 16#37#;
	WM_IME_STARTCOMPOSITION  : constant := 16#010D#;
	WM_IME_ENDCOMPOSITION    : constant := 16#010E#;
	WM_IME_COMPOSITION       : constant := 16#010F#;
	WM_IME_KEYLAST           : constant := 16#010F#;
	WM_IME_SETCONTEXT        : constant := 16#0281#;
	WM_IME_NOTIFY            : constant := 16#0282#;
	WM_IME_CONTROL           : constant := 16#0283#;
	WM_IME_COMPOSITIONFULL   : constant := 16#0284#;
	WM_IME_SELECT            : constant := 16#0285#;
	WM_IME_CHAR              : constant := 16#0286#;
	WM_IME_REQUEST           : constant := 16#0288#;
	WM_IME_KEYDOWN           : constant := 16#0290#;
	WM_IME_KEYUP             : constant := 16#0291#;
	WM_MDICREATE             : constant := 16#0220#;
	WM_MDIDESTROY            : constant := 16#0221#;
	WM_MDIACTIVATE           : constant := 16#0222#;
	WM_MDIRESTORE            : constant := 16#0223#;
	WM_MDINEXT               : constant := 16#0224#;
	WM_MDIMAXIMIZE           : constant := 16#0225#;
	WM_MDITILE               : constant := 16#0226#;
	WM_MDICASCADE            : constant := 16#0227#;
	WM_MDIICONARRANGE        : constant := 16#0228#;
	WM_MDIGETACTIVE          : constant := 16#0229#;
	WM_MDISETMENU            : constant := 16#0230#;
	WM_ENTERSIZEMOVE         : constant := 16#0231#;
	WM_EXITSIZEMOVE          : constant := 16#0232#;
	WM_DROPFILES             : constant := 16#0233#;
	WM_MDIREFRESHMENU        : constant := 16#0234#;
	WM_MOUSEHOVER            : constant := 16#02A1#;
	WM_MOUSELEAVE            : constant := 16#02A3#;
	WM_CUT                   : constant := 16#0300#;
	WM_COPY                  : constant := 16#0301#;
	WM_PASTE                 : constant := 16#0302#;
	WM_CLEAR                 : constant := 16#0303#;
	WM_UNDO                  : constant := 16#0304#;
	WM_RENDERFORMAT          : constant := 16#0305#;
	WM_RENDERALLFORMATS      : constant := 16#0306#;
	WM_DESTROYCLIPBOARD      : constant := 16#0307#;
	WM_DRAWCLIPBOARD         : constant := 16#0308#;
	WM_PAINTCLIPBOARD        : constant := 16#0309#;
	WM_VSCROLLCLIPBOARD      : constant := 16#030A#;
	WM_SIZECLIPBOARD         : constant := 16#030B#;
	WM_ASKCBFORMATNAME       : constant := 16#030C#;
	WM_CHANGECBCHAIN         : constant := 16#030D#;
	WM_HSCROLLCLIPBOARD      : constant := 16#030E#;
	WM_QUERYNEWPALETTE       : constant := 16#030F#;
	WM_PALETTEISCHANGING     : constant := 16#0310#;
	WM_PALETTECHANGED        : constant := 16#0311#;
	WM_HOTKEY                : constant := 16#0312#;
	WM_PRINT                 : constant := 16#0317#;
	WM_PRINTCLIENT           : constant := 16#0318#;
	WM_HANDHELDFIRST         : constant := 16#0358#;
	WM_HANDHELDLAST          : constant := 16#035F#;
	WM_PENWINFIRST           : constant := 16#0380#;
	WM_PENWINLAST            : constant := 16#038F#;
	WM_COALESCE_FIRST        : constant := 16#0390#;
	WM_COALESCE_LAST         : constant := 16#039F#;
	WM_DDE_FIRST             : constant := 16#03E0#;
	WM_DDE_INITIATE          : constant := WM_DDE_FIRST + 0;
	WM_DDE_TERMINATE         : constant := WM_DDE_FIRST + 1;
	WM_DDE_ADVISE            : constant := WM_DDE_FIRST + 2;
	WM_DDE_UNADVISE          : constant := WM_DDE_FIRST + 3;
	WM_DDE_ACK               : constant := WM_DDE_FIRST + 4;
	WM_DDE_DATA              : constant := WM_DDE_FIRST + 5;
	WM_DDE_REQUEST           : constant := WM_DDE_FIRST + 6;
	WM_DDE_POKE              : constant := WM_DDE_FIRST + 7;
	WM_DDE_EXECUTE           : constant := WM_DDE_FIRST + 8;
	WM_DDE_LAST              : constant := WM_DDE_FIRST + 8;
	WM_USER                  : constant := 16#0400#;
	WM_APP                   : constant := 16#8000#;

	SC_SIZE : constant := 61440;
	SC_MOVE : constant := 61456;
	SC_MINIMIZE : constant := 61472;
	SC_MAXIMIZE : constant := 61488;
	SC_NEXTWINDOW : constant := 61504;
	SC_PREVWINDOW : constant := 61520;
	SC_CLOSE : constant := 61536;
	SC_VSCROLL : constant := 61552;
	SC_HSCROLL : constant := 61568;
	SC_MOUSEMENU : constant := 61584;
	SC_KEYMENU : constant := 61696;
	SC_ARRANGE : constant := 61712;
	SC_RESTORE : constant := 61728;
	SC_TASKLIST : constant := 61744;
	SC_SCREENSAVE : constant := 61760;
	SC_HOTKEY : constant := 61776;
	SC_DEFAULT : constant := 61792;
	SC_MONITORPOWER : constant := 61808;
	SC_CONTEXTHELP : constant := 61824;
	SC_SEPARATOR : constant := 61455;

	WS_OVERLAPPED          : constant := 16#00000000#;
	WS_POPUP               : constant := 16#80000000#;
	WS_CHILD               : constant := 16#40000000#;
	WS_MINIMIZE            : constant := 16#20000000#;
	WS_VISIBLE             : constant := 16#10000000#;
	WS_DISABLED            : constant := 16#08000000#;
	WS_CLIPSIBLINGS        : constant := 16#04000000#;
	WS_CLIPCHILDREN        : constant := 16#02000000#;
	WS_MAXIMIZE            : constant := 16#01000000#;
	WS_CAPTION             : constant := 16#00c00000#;
	WS_BORDER              : constant := 16#00800000#;
	WS_DLGFRAME            : constant := 16#00400000#;
	WS_VSCROLL             : constant := 16#00200000#;
	WS_HSCROLL             : constant := 16#00100000#;
	WS_SYSMENU             : constant := 16#00080000#;
	WS_THICKFRAME          : constant := 16#00040000#;
	WS_GROUP               : constant := 16#00020000#;
	WS_TABSTOP             : constant := 16#00010000#;
	WS_MINIMIZEBOX         : constant := 16#00020000#;
	WS_MAXIMIZEBOX         : constant := 16#00010000#;
	WS_TILED               : constant := 16#00000000#;
	WS_ICONIC              : constant := 16#20000000#;
	WS_SIZEBOX             : constant := 16#00040000#;
	WS_TILEDWINDOW         : constant := 16#00cf0000#;
	WS_OVERLAPPEDWINDOW    : constant := 16#00cf0000#;
	WS_POPUPWINDOW         : constant := 16#80880000#;
	WS_CHILDWINDOW         : constant := 16#40000000#;

	ES_LEFT                : constant := 16#0#;
	ES_CENTER              : constant := 16#1#;
	ES_RIGHT               : constant := 16#2#;
	ES_MULTILINE           : constant := 16#4#;
	ES_UPPERCASE           : constant := 16#8#;
	ES_LOWERCASE           : constant := 16#10#;
	ES_PASSWORD            : constant := 16#20#;
	ES_AUTOVSCROLL         : constant := 16#40#;
	ES_AUTOHSCROLL         : constant := 16#80#;
	ES_NOHIDESEL           : constant := 16#100#;
	ES_OEMCONVERT          : constant := 16#400#;
	ES_READONLY            : constant := 16#800#;
	ES_WANTRETURN          : constant := 16#1000#;

	WS_EX_DLGMODALFRAME    : constant := 16#00000001#;
	WS_EX_NOPARENTNOTIFY   : constant := 16#00000004#;
	WS_EX_TOPMOST          : constant := DWORD'(16#00000008#);
	WS_EX_ACCEPTFILES      : constant := 16#00000010#;
	WS_EX_TRANSPARENT      : constant := 16#00000020#;
	WS_EX_MDICHILD         : constant := 16#00000040#;
	WS_EX_TOOLWINDOW       : constant := 16#00000080#;
	WS_EX_WINDOWEDGE       : constant := DWORD'(16#00000100#);
	WS_EX_CLIENTEDGE       : constant := DWORD'(16#00000200#);
	WS_EX_CONTEXTHELP      : constant := 16#00000400#;
	WS_EX_RIGHT            : constant := 16#00001000#;
	WS_EX_LEFT             : constant := 0;
	WS_EX_RTLREADING       : constant := 16#00002000#;
	WS_EX_LTRREADING       : constant := 0;
	WS_EX_LEFTSCROLLBAR    : constant := 16#00004000#;
	WS_EX_RIGHTSCROLLBAR   : constant := 0;
	WS_EX_CONTROLPARENT    : constant := 16#00010000#;
	WS_EX_STATICEDGE       : constant := 16#00020000#;
	WS_EX_APPWINDOW        : constant := 16#00040000#;
	WS_EX_OVERLAPPEDWINDOW : constant := DWORD'(WS_EX_WINDOWEDGE or WS_EX_CLIENTEDGE);
	WS_EX_PALETTEWINDOW    : constant := DWORD'(WS_EX_WINDOWEDGE or WS_EX_TOOLWINDOW or WS_EX_TOPMOST);
	WS_EX_LAYERED          : constant := 16#00080000#;
	WS_EX_NOINHERITLAYOUT  : constant := 16#00100000#;
	WS_EX_LAYOUTRTL        : constant := 16#00400000#;
	WS_EX_NOACTIVATE       : constant := 16#08000000#;

	type WNDPROC is access function (Wnd: HWND; Msg: UINT; W: WPARAM; L: LPARAM) return LRESULT;
	pragma Convention(Stdcall, WNDPROC);

	type WNDCLASSA is      
		record
			style        : UINT;
			lpfnWndProc  : WNDPROC;
			cbClsExtra   : INT;
			cbWndExtra   : INT;
			hInstance    : Windows.HINSTANCE;
			hIcon        : Windows.HICON;
			hCursor      : Windows.HCURSOR;
			hbrBackground: HBRUSH;
			lpszMenuName : PCASTR;
			lpszClassName: PCASTR;
		end record;
	pragma Convention(C, WNDCLASSA);
	type WNDCLASSW is
		record
			style        : UINT;
			lpfnWndProc  : WNDPROC;
			cbClsExtra   : INT;
			cbWndExtra   : INT;
			hInstance    : Windows.HINSTANCE;
			hIcon        : Windows.HICON;
			hCursor      : Windows.HCURSOR;
			hbrBackground: HBRUSH;
			lpszMenuName : PCWSTR;
			lpszClassName: PCWSTR;
		end record;
	pragma Convention(C, WNDCLASSW);
	subtype WNDCLASS is WNDCLASSA;

	type MSG is
		record
			hwnd   : Windows.HWND;
			message: UINT;
			wParam : Windows.WPARAM;
			lParam : Windows.LPARAM;
			time   : DWORD;
			pt     : POINT;
		end record;
	pragma Convention(C, MSG);

	function RegisterClassA(WndClass: access WNDCLASSA) return ATOM;
	function RegisterClassW(WndClass: access WNDCLASSW) return ATOM;
	function RegisterClass (WndClass: access WNDCLASSA) return ATOM renames RegisterClassA;
	pragma Import(stdcall, RegisterClassA, "RegisterClassA");
	pragma Import(stdcall, RegisterClassW, "RegisterClassW");

	CW_USEDEFAULT : constant INT := INT'First;

	function CreateWindowExA(ExStyle: DWORD; ClassName: PCASTR; Text: PCASTR; Style: DWORD; X, Y, Width, Height: INT; WndParent: HWND; Menu: HMENU; Instance: HINSTANCE; Param: Pointer) return HWND;
	function CreateWindowExW(ExStyle: DWORD; ClassName: PCWSTR; Text: PCWSTR; Style: DWORD; X, Y, Width, Height: INT; WndParent: HWND; Menu: HMENU; Instance: HINSTANCE; Param: Pointer) return HWND;
	function CreateWindowEx (ExStyle: DWORD; ClassName: PCTSTR; Text: PCTSTR; Style: DWORD; X, Y, Width, Height: INT; WndParent: HWND; Menu: HMENU; Instance: HINSTANCE; Param: Pointer) return HWND renames CreateWindowExA;
	pragma Import(stdcall, CreateWindowExA, "CreateWindowExA");
	pragma Import(stdcall, CreateWindowExW, "CreateWindowExW");
	function CreateWindowA(ClassName: PCASTR; Text: PCASTR; Style: DWORD; X, Y, Width, Height: INT; WndParent: HWND; Menu: HMENU; Instance: HINSTANCE; Param: Pointer) return HWND;
	function CreateWindowW(ClassName: PCWSTR; Text: PCWSTR; Style: DWORD; X, Y, Width, Height: INT; WndParent: HWND; Menu: HMENU; Instance: HINSTANCE; Param: Pointer) return HWND;
	function CreateWindow (ClassName: PCTSTR; Text: PCTSTR; Style: DWORD; X, Y, Width, Height: INT; WndParent: HWND; Menu: HMENU; Instance: HINSTANCE; Param: Pointer) return HWND renames CreateWindowA;
	pragma Inline(CreateWindowA);
	pragma Inline(CreateWindowW);

	function DefWindowProcA(Wnd: HWND; Msg: UINT; W: WPARAM; L: LPARAM) return LRESULT;
	function DefWindowProcW(Wnd: HWND; Msg: UINT; W: WPARAM; L: LPARAM) return LRESULT;
	function DefWindowProc (Wnd: HWND; Msg: UINT; W: WPARAM; L: LPARAM) return LRESULT renames DefWindowProcA;
	pragma Import(stdcall, DefWindowProcA, "DefWindowProcA");
	pragma Import(stdcall, DefWindowProcW, "DefWindowProcW");

	procedure PostQuitMessage(ExitCode: INT);
	pragma Import(stdcall, PostQuitMessage, "PostQuitMessage");

	function ShowWindow(Wnd: HWND; CmdShow: INT) return BOOL;
	pragma Import(stdcall, ShowWindow, "ShowWindow");
	procedure ShowWindow(Wnd: HWND; CmdShow: INT);
	pragma Inline(ShowWindow);

	function UpdateWindow(Wnd: HWND) return BOOL;
	pragma Import(stdcall, UpdateWindow, "UpdateWindow");
	procedure UpdateWindow(Wnd: HWND);
	pragma Inline(UpdateWindow);

	function SetFocus(Wnd: HWND) return HWND;
	pragma Import(stdcall, SetFocus, "SetFocus");
	procedure SetFocus(Wnd: HWND);
	pragma Inline(SetFocus);

	function MoveWindow(Wnd: HWND; X, Y, Width, Height: INT; Repaint: BOOL) return BOOL;
	pragma Import(stdcall, MoveWindow, "MoveWindow");
	procedure MoveWindow(Wnd: HWND; X, Y, Width, Height: INT; Repaint: BOOL);
	pragma Inline(MoveWindow);

	function SetWindowTextA(Wnd: HWND; String: PCASTR) return BOOL;
	function SetWindowTextW(Wnd: HWND; String: PCWSTR) return BOOL;
	function SetWindowText (Wnd: HWND; String: PCTSTR) return BOOL renames SetWindowTextA;
	pragma Import(stdcall, SetWindowTextA, "SetWindowTextA");
	pragma Import(stdcall, SetWindowTextW, "SetWindowTextW");
	procedure SetWindowTextA(Wnd: HWND; String: PCASTR);
	procedure SetWindowTextW(Wnd: HWND; String: PCWSTR);
	procedure SetWindowText (Wnd: HWND; String: PCTSTR) renames SetWindowTextA;
	pragma Inline(SetWindowTextA);
	pragma Inline(SetWindowTextW);

	function GetWindowTextA(Wnd: HWND; String: PASTR; MaxCount: INT) return INT;
	function GetWindowTextW(Wnd: HWND; String: PWSTR; MaxCount: INT) return INT;
	function GetWindowText (Wnd: HWND; String: PTSTR; MaxCount: INT) return INT renames GetWindowTextA;
	pragma Import(stdcall, GetWindowTextA, "GetWindowTextA");
	pragma Import(stdcall, GetWindowTextW, "GetWindowTextW");

	function GetWindowTextLengthA(Wnd: HWND) return INT;
	function GetWindowTextLengthW(Wnd: HWND) return INT;
	function GetWindowTextLength (Wnd: HWND) return INT renames GetWindowTextLengthA;
	pragma Import(stdcall, GetWindowTextLengthA, "GetWindowTextLengthA");
	pragma Import(stdcall, GetWindowTextLengthW, "GetWindowTextLengthW");

	function GetClientRect(Wnd: HWND; PRect: access RECT) return BOOL;
	pragma Import(stdcall, GetClientRect, "GetClientRect");
	procedure GetClientRect(Wnd: HWND; PRect: access RECT);
	pragma Inline(GetClientRect);

	GWL_WNDPROC            : constant := -4;
	GWL_HINSTANCE          : constant := -6;
	GWL_HWNDPARENT         : constant := -8;
	GWL_STYLE              : constant := -16;
	GWL_EXSTYLE            : constant := -20;
	GWL_USERDATA           : constant := -21;
	GWL_ID                 : constant := -12;

	function GetWindowLongA(Wnd: HWND; Index: INT) return LONG;
	function GetWindowLongW(Wnd: HWND; Index: INT) return LONG;
	function GetWindowLong (Wnd: HWND; Index: INT) return LONG renames GetWindowLongA;
	pragma Import(stdcall, GetWindowLongA, "GetWindowLongA");
	pragma Import(stdcall, GetWindowLongW, "GetWindowLongW");

	function SetWindowLongA(Wnd: HWND; Index: INT; NewLong: LONG) return LONG;
	function SetWindowLongW(Wnd: HWND; Index: INT; NewLong: LONG) return LONG;
	function SetWindowLong (Wnd: HWND; Index: INT; NewLong: LONG) return LONG renames SetWindowLongA;
	pragma Import(stdcall, SetWindowLongA, "SetWindowLongA");
	pragma Import(stdcall, SetWindowLongW, "SetWindowLongW");
	procedure SetWindowLongA(Wnd: HWND; Index: INT; NewLong: LONG);
	procedure SetWindowLongW(Wnd: HWND; Index: INT; NewLong: LONG);
	procedure SetWindowLong (Wnd: HWND; Index: INT; NewLong: LONG) renames SetWindowLongA;
	pragma Inline(SetWindowLongA);
	pragma Inline(SetWindowLongW);
	
	type WNDENUMPROC is access function(Wnd : HWND; Param : LPARAM) return BOOL;
	pragma Convention(stdcall, WNDENUMPROC);
	function EnumWindows(EnumFunc: WNDENUMPROC; Param: LPARAM) return BOOL;
	pragma Import(stdcall, EnumWindows, "EnumWindows");
	
	generic
		with function Each_Window(Window : HWND) return Boolean;
	function Enumerate_Windows return HWND;
	
	function PostMessageA(Wnd: HWND; Message: UINT; W: WPARAM; L: LPARAM) return BOOL;
	function PostMessageW(Wnd: HWND; Message: UINT; W: WPARAM; L: LPARAM) return BOOL;
	function PostMessage (Wnd: HWND; Message: UINT; W: WPARAM; L: LPARAM) return BOOL renames PostMessageA;
	pragma Import(stdcall, PostMessageA, "PostMessageA");
	pragma Import(stdcall, PostMessageW, "PostMessageW");
	procedure PostMessageA(Wnd: HWND; Message: UINT; W: WPARAM; L: LPARAM);
	procedure PostMessageW(Wnd: HWND; Message: UINT; W: WPARAM; L: LPARAM);
	procedure PostMessage (Wnd: HWND; Message: UINT; W: WPARAM; L: LPARAM) renames PostMessageA;
	pragma Inline(PostMessageA);
	pragma Inline(PostMessageW);

	function GetMessageA(Message: access MSG; Wnd: HWND; MsgFilterMin, MsgFilterMax: UINT) return BOOL;
	function GetMessageW(Message: access MSG; Wnd: HWND; MsgFilterMin, MsgFilterMax: UINT) return BOOL;
	function GetMessage (Message: access MSG; Wnd: HWND; MsgFilterMin, MsgFilterMax: UINT) return BOOL renames GetMessageA;
	pragma Import(stdcall, GetMessageA, "GetMessageA");
	pragma Import(stdcall, GetMessageW, "GetMessageW");

	function TranslateMessage(Message: access MSG) return BOOL;
	pragma Import(stdcall, TranslateMessage, "TranslateMessage");
	procedure TranslateMessage(Message: access MSG);
	pragma Inline(TranslateMessage);

	function DispatchMessageA(Message: access MSG) return LRESULT;
	function DispatchMessageW(Message: access MSG) return LRESULT;
	function DispatchMessage (Message: access MSG) return LRESULT renames DispatchMessageA;
	pragma Import(stdcall, DispatchMessageA, "DispatchMessageA");
	pragma Import(stdcall, DispatchMessageW, "DispatchMessageW");
	procedure DispatchMessageA(Message: access MSG);
	procedure DispatchMessageW(Message: access MSG);
	procedure DispatchMessage (Message: access MSG) renames DispatchMessageA;
	pragma Inline(DispatchMessageA);
	pragma Inline(DispatchMessageW);

	IDI_APPLICATIONA : constant PCASTR := PCASTR(Address_To_PCASTR_Conversions.To_Pointer(System.Storage_Elements.To_Address(32512)));
	IDI_APPLICATIONW : constant PCWSTR := PCWSTR(Address_To_PCWSTR_Conversions.To_Pointer(System.Storage_Elements.To_Address(32512)));
	IDI_APPLICATION  : PCTSTR renames IDI_APPLICATIONA;

	function LoadIconA(Instance: HINSTANCE; IconName: PCASTR) return HICON;
	function LoadIconW(Instance: HINSTANCE; IconName: PCWSTR) return HICON;
	function LoadIcon (Instance: HINSTANCE; IconName: PCTSTR) return HICON renames LoadIconA;
	pragma Import(stdcall, LoadIconA, "LoadIconA");
	pragma Import(stdcall, LoadIconW, "LoadIconW");

	IDC_ARROWA : constant PCASTR := PCASTR(Address_To_PCASTR_Conversions.To_Pointer(System.Storage_Elements.To_Address(32512)));
	IDC_ARROWW : constant PCWSTR := PCWSTR(Address_To_PCWSTR_Conversions.To_Pointer(System.Storage_Elements.To_Address(32512)));
	IDC_ARROW  : PCTSTR renames IDC_ARROWA;

	function LoadCursorA(Instance: HINSTANCE; CursorName: PCASTR) return HCursor;
	function LoadCursorW(Instance: HINSTANCE; CursorName: PCWSTR) return HCursor;
	function LoadCursor (Instance: HINSTANCE; CursorName: PCTSTR) return HCursor renames LoadCursorA;
	pragma Import(stdcall, LoadCursorA, "LoadCursorA");
	pragma Import(stdcall, LoadCursorW, "LoadCursorW");

	MF_INSERT              : constant := 16#0000#;
	MF_CHANGE              : constant := 16#0080#;
	MF_APPEND              : constant := 16#0100#;
	MF_DELETE              : constant := 16#0200#;
	MF_REMOVE              : constant := 16#1000#;
	MF_BYCOMMAND           : constant := 16#0000#;
	MF_BYPOSITION          : constant := 16#0400#;
	MF_SEPARATOR           : constant := 16#0800#;
	MF_ENABLED             : constant := 16#0000#;
	MF_GRAYED              : constant := 16#0001#;
	MF_DISABLED            : constant := 16#0002#;
	MF_UNCHECKED           : constant := 16#0000#;
	MF_CHECKED             : constant := 16#0008#;
	MF_USECHECKBITMAPS     : constant := 16#0200#;
	MF_STRING              : constant := 16#0000#;
	MF_BITMAP              : constant := 16#0004#;
	MF_OWNERDRAW           : constant := 16#0100#;
	MF_POPUP               : constant := 16#0010#;
	MF_MENUBARBREAK        : constant := 16#0020#;
	MF_MENUBREAK           : constant := 16#0040#;
	MF_UNHILITE            : constant := 16#0000#;
	MF_HILITE              : constant := 16#0080#;
	MF_SYSMENU             : constant := 16#2000#;
	MF_HELP                : constant := 16#4000#;
	MF_MOUSESELECT         : constant := 16#8000#;
	MF_END                 : constant := 16#0080#;

	function CreateMenu return HMENU;
	pragma Import(stdcall, CreateMenu, "CreateMenu");

	function CreatePopupMenu return HMENU;
	pragma Import(stdcall, CreatePopupMenu, "CreatePopupMenu");

	function DestroyMenu(Menu: HMENU) return BOOL;
	pragma Import(stdcall, DestroyMenu, "DestroyMenu");
	procedure DestroyMenu(Menu: HMENU);
	pragma Inline(DestroyMenu);

	function AppendMenuA(Menu: HMENU; Flags: UINT; IDNewItem: UINT; NewItem: PCASTR) return BOOL;
	function AppendMenuW(Menu: HMENU; Flags: UINT; IDNewItem: UINT; NewItem: PCWSTR) return BOOL;
	function AppendMenu (Menu: HMENU; Flags: UINT; IDNewItem: UINT; NewItem: PCTSTR) return BOOL renames AppendMenuA;
	pragma Import(stdcall, AppendMenuA, "AppendMenuA");
	pragma Import(stdcall, AppendMenuW, "AppendMenuW");
	procedure AppendMenuA(Menu: HMENU; Flags: UINT; IDNewItem: UINT; NewItem: PCASTR);
	procedure AppendMenuW(Menu: HMENU; Flags: UINT; IDNewItem: UINT; NewItem: PCWSTR);
	procedure AppendMenu (Menu: HMENU; Flags: UINT; IDNewItem: UINT; NewItem: PCTSTR) renames AppendMenuA;
	pragma Inline(AppendMenuA);
	pragma Inline(AppendMenuW);

	function SetMenu(Wnd: HWND; Menu: HMENU) return BOOL;
	pragma Import(stdcall, SetMenu, "SetMenu");
	procedure SetMenu(Wnd: HWND; Menu: HMENU);
	pragma Inline(SetMenu);

	MB_OK                  : constant := 16#00000000#;
	MB_OKCANCEL            : constant := 16#00000001#;
	MB_ABORTRETRYIGNORE    : constant := 16#00000002#;
	MB_YESNOCANCEL         : constant := 16#00000003#;
	MB_YESNO               : constant := 16#00000004#;
	MB_RETRYCANCEL         : constant := 16#00000005#;
	MB_ICONHAND            : constant := 16#00000010#;
	MB_ICONQUESTION        : constant := 16#00000020#;
	MB_ICONEXCLAMATION     : constant := 16#00000030#;
	MB_ICONASTERISK        : constant := 16#00000040#;
	MB_ICONINFORMATION     : constant := 16#00000040#;
	MB_ICONSTOP            : constant := 16#00000010#;
	MB_DEFBUTTON1          : constant := 16#00000000#;
	MB_DEFBUTTON2          : constant := 16#00000100#;
	MB_DEFBUTTON3          : constant := 16#00000200#;
	MB_APPLMODAL           : constant := 16#00000000#;
	MB_SYSTEMMODAL         : constant := 16#00001000#;
	MB_TASKMODAL           : constant := 16#00002000#;
	MB_NOFOCUS             : constant := 16#00008000#;
	MB_SETFOREGROUND       : constant := 16#00010000#;
	MB_DEFAULT_DESKTOP_ONLY: constant := 16#00020000#;
	MB_SERVICE_NOTIFICATION: constant := 16#00040000#;
	MB_TYPEMASK            : constant := 16#0000000f#;
	MB_ICONMASK            : constant := 16#000000f0#;
	MB_DEFMASK             : constant := 16#00000f00#;
	MB_MODEMASK            : constant := 16#00003000#;
	MB_MISCMASK            : constant := 16#0000c000#;

	function MessageBoxA(Wnd: HWND; Text, Caption: PCASTR; uType: UINT) return INT;
	function MessageBoxW(Wnd: HWND; Text, Caption: PCWSTR; uType: UINT) return INT;
	function MessageBox (Wnd: HWND; Text, Caption: PCTSTR; uType: UINT) return INT renames MessageBoxA;
	pragma Import(stdcall, MessageBoxA, "MessageBoxA");
	pragma Import(stdcall, MessageBoxW, "MessageBoxW");
	procedure MessageBoxA(Wnd: HWND; Text, Caption: PCASTR; uType: UINT);
	procedure MessageBoxW(Wnd: HWND; Text, Caption: PCWSTR; uType: UINT);
	procedure MessageBox (Wnd: HWND; Text, Caption: PCTSTR; uType: UINT) renames MessageBoxA;
	pragma Inline(MessageBoxA);
	pragma Inline(MessageBoxW);

	VK_LBUTTON : constant := 1;
	VK_RBUTTON : constant := 2;
	VK_CANCEL : constant := 3;
	VK_MBUTTON : constant := 4;  -- NOT contiguous with L & RBUTTON 
	VK_BACK : constant := 8;
	VK_TAB : constant := 9;
	VK_CLEAR : constant := 12;
	VK_RETURN : constant := 13;
	VK_SHIFT : constant := 16;
	VK_CONTROL : constant := 17;
	VK_MENU : constant := 18;
	VK_PAUSE : constant := 19;
	VK_CAPITAL : constant := 20;
	VK_KANA : constant := 21;
	VK_HANGUL : constant := 21;
	VK_JUNJA : constant := 23;
	VK_FINAL : constant := 24;
	VK_HANJA : constant := 25;
	VK_KANJI : constant := 25;
	VK_CONVERT : constant := 28;
	VK_NONCONVERT : constant := 29;
	VK_ACCEPT : constant := 30;
	VK_MODECHANGE : constant := 31;
	VK_ESCAPE : constant := 27;
	VK_SPACE : constant := 32;
	VK_PRIOR : constant := 33;
	VK_NEXT : constant := 34;
	VK_END : constant := 35;
	VK_HOME : constant := 36;
	VK_LEFT : constant := 37;
	VK_UP : constant := 38;
	VK_RIGHT : constant := 39;
	VK_DOWN : constant := 40;
	VK_SELECT : constant := 41;
	VK_PRINT : constant := 42;
	VK_EXECUTE : constant := 43;
	VK_SNAPSHOT : constant := 44;
	VK_INSERT : constant := 45;
	VK_DELETE : constant := 46;
	VK_HELP : constant := 47;
	-- VK_0 thru VK_9 are the same as ASCII '0' thru '9' ($30 - $39)
	-- VK_A thru VK_Z are the same as ASCII 'A' thru 'Z' ($41 - $5A)
	VK_LWIN : constant := 91;
	VK_RWIN : constant := 92;
	VK_APPS : constant := 93;
	VK_NUMPAD0 : constant := 96;
	VK_NUMPAD1 : constant := 97;
	VK_NUMPAD2 : constant := 98;
	VK_NUMPAD3 : constant := 99;
	VK_NUMPAD4 : constant := 100;
	VK_NUMPAD5 : constant := 101;
	VK_NUMPAD6 : constant := 102;
	VK_NUMPAD7 : constant := 103;
	VK_NUMPAD8 : constant := 104;
	VK_NUMPAD9 : constant := 105;
	VK_MULTIPLY : constant := 106;
	VK_ADD : constant := 107;
	VK_SEPARATOR : constant := 108;
	VK_SUBTRACT : constant := 109;
	VK_DECIMAL : constant := 110;
	VK_DIVIDE : constant := 111;
	VK_F1 : constant := 112;
	VK_F2 : constant := 113;
	VK_F3 : constant := 114;
	VK_F4 : constant := 115;
	VK_F5 : constant := 116;
	VK_F6 : constant := 117;
	VK_F7 : constant := 118;
	VK_F8 : constant := 119;
	VK_F9 : constant := 120;
	VK_F10 : constant := 121;
	VK_F11 : constant := 122;
	VK_F12 : constant := 123;
	VK_F13 : constant := 124;
	VK_F14 : constant := 125;
	VK_F15 : constant := 126;
	VK_F16 : constant := 127;
	VK_F17 : constant := 128;
	VK_F18 : constant := 129;
	VK_F19 : constant := 130;
	VK_F20 : constant := 131;
	VK_F21 : constant := 132;
	VK_F22 : constant := 133;
	VK_F23 : constant := 134;
	VK_F24 : constant := 135;
	VK_NUMLOCK : constant := 144;
	VK_SCROLL : constant := 145;
	-- VK_L & VK_R - left and right Alt, Ctrl and Shift virtual keys.
	-- Used only as parameters to GetAsyncKeyState() and GetKeyState().
	-- No other API or message will distinguish left and right keys in this way. 
	VK_LSHIFT : constant := 160;
	VK_RSHIFT : constant := 161;
	VK_LCONTROL : constant := 162;
	VK_RCONTROL : constant := 163;
	VK_LMENU : constant := 164;
	VK_RMENU : constant := 165;
	VK_PROCESSKEY : constant := 229;
	VK_ATTN : constant := 246;
	VK_CRSEL : constant := 247;
	VK_EXSEL : constant := 248;
	VK_EREOF : constant := 249;
	VK_PLAY : constant := 250;
	VK_ZOOM : constant := 251;
	VK_NONAME : constant := 252;
	VK_PA1 : constant := 253;
	VK_OEM_CLEAR : constant := 254;

	function GetKeyState(VirtKey: INT) return WORD; -- SHORT
	pragma Import(stdcall, GetKeyState, "GetKeyState");

---- WINGDI.h -----

	WHITE_BRUSH                : constant := 0; 
	LTGRAY_BRUSH               : constant := 1; 
	GRAY_BRUSH                 : constant := 2; 
	DKGRAY_BRUSH               : constant := 3; 
	BLACK_BRUSH                : constant := 4; 
	NULL_BRUSH                 : constant := 5; 
	HOLLOW_BRUSH               : constant := 5; 
	WHITE_PEN                  : constant := 6; 
	BLACK_PEN                  : constant := 7; 
	NULL_PEN                   : constant := 8; 
	OEM_FIXED_FONT             : constant := 10;
	ANSI_FIXED_FONT            : constant := 11;
	ANSI_VAR_FONT              : constant := 12;
	SYSTEM_FONT                : constant := 13;
	DEVICE_DEFAULT_FONT        : constant := 14;
	DEFAULT_PALETTE            : constant := 15;
	SYSTEM_FIXED_FONT          : constant := 16;
	STOCK_LAST                 : constant := 16;

	function GetStockObject(Object: INT) return HGDIOBJ;
	pragma Import(stdcall, GetStockObject, "GetStockObject");

---- ADVAPI.h ----

	function GetUserNameA(Buffer: PASTR; Size: access DWORD) return BOOL;
	function GetUserNameW(Buffer: PWSTR; Size: access DWORD) return BOOL;
	function GetUserName (Buffer: PTSTR; Size: access DWORD) return BOOL renames GetUserNameA;
	pragma Import(stdcall, GetUserNameA, "GetUserNameA");
	pragma Import(stdcall, GetUserNameW, "GetUserNameW");

---- utilities ----

	function Retrieve_hInstance return HINSTANCE;
	pragma Import(C, Retrieve_hInstance, "rts_get_hInstance");
	function Retrieve_hPrevInstance return HINSTANCE;
	pragma Import(C, Retrieve_hPrevInstance, "rts_get_hPrevInstance");
	function Retrieve_lpCmdline return PCASTR;
	pragma Import(C, Retrieve_lpCmdline,  "rts_get_lpCommandLine");
	function Retrieve_nCmdShow return INT;
	pragma Import(C, Retrieve_nCmdShow, "rts_get_nShowCmd");

	function Address_Of(S : in String) return PASTR;
	pragma inline(Address_Of);
	function Address_Of(S : in String) return PCASTR;
	pragma inline(Address_Of);
	function Address_Of(S : in Wide_String) return PWSTR;
	pragma inline(Address_Of);
	function Address_Of(S : in Wide_String) return PCWSTR;
	pragma inline(Address_Of);
	-- Ada05:
	-- function Address_Of(S : not null access String) return not null access ACHAR;
	-- function Address_Of(S : not null access Wide_String) return not null access WCHAR;
	-- function Address_Of(S : not null access constant String) return not null access constant ACHAR;
	-- function Address_Of(S : not null access constant Wide_String) return not null access constant WCHAR;

	-- Ada05:
	function To_String(S : not null access constant ACHAR) return String;
	function To_String(S : not null access constant WCHAR) return Wide_String;

	procedure Assign(D : out ASTR; S : in String);
	procedure Assign(D : out WSTR; S : in Wide_String);

	function Error_Message(Code : DWORD) return String;

private

	procedure memmove(dest, src : Pointer; count : Interfaces.C.size_t);
	pragma Import(C, memmove, "memmove");
	
end Windows;
